#Created by Jack Mumford in 2021 - Version 0.4
#2021, Jack Mumford.

Clear-Host
$script:OutputData = $null
$script:OutputFilePath = $null
$script:SiteID = $null
$script:SitePermissions = @()
$script:CSVListPermissions = @()
$script:SiteName = $null
$script:OutputFolderLocation = $null

function GetListPermissions{
    
    Write-Host "Getting List Permissions"

    try{
    
        #Get all lists on the site
        $Lists = Get-PnPList -Includes BaseTemplate,HasUniqueRoleAssignments

        #Begin the ListPermission part within the CSV
        $script:CSVListPermissions += @("`n`nList Permissions,Permission Level,Entity Type")
        $script:CSVListPermissions += @(",,")

    $ExcludedLists = @("App Packages","appdata","appfiles","Apps in Testing","Cache Profiles","Composed Looks","Content and Structure Reports","Content type publishing error log","Converted Forms",
     "Device Channels","Form Templates","fpdatasources","Get started with Apps for Office and SharePoint","List Template Gallery", "Long Running Operation Status","Maintenance Log Library", "Style Library",
     ,"Master Docs","Master Page Gallery","MicroFeed","NintexFormXml","Quick Deploy Items","Relationships List","Reusable Content","Search Config List", "Solution Gallery", "Site Collection Images",
     "Suggested Content Browser Locations","TaxonomyHiddenList","User Information List","Web Part Gallery","wfpub","wfsvc","Workflow History","Workflow Tasks", "Preservation Hold Library")

        foreach($List in $Lists){
            
            if($List.BaseTemplate -eq "101" -or $List.BaseTemplate -eq "100"){
                
                if($List.HasUniqueRoleAssignments -eq $false -and $ExcludedLists -notcontains $List.Title){
                    #Doesnt have unique perms, but it could contain and item that has unique perms!
                    $ListName = $List.Title
                    $ListBaseType = $List.BaseType
                    $script:CSVListPermissions += @("'$ListName' - $ListBaseType,Inheriting from Root site,This list could contain unique items")
                    $script:CSVListPermissions += @(",,")

                }
                elseif($List.Title -notcontains "Style Library" -and $ExcludedLists -notcontains $List.Title){
                
                    #Has unique perms, check if any sub items have unique items to its parent
                    $ListName = $List.Title
                    $ListBaseType = $List.BaseType
                    $script:CSVListPermissions += @("'$ListName' - $ListBaseType,,This list could contain unique items")
                    Write-Host $List.Title

                    #Get all lists on the site
                    $ListPermissions = Get-PnPList -Identity $List.Id -Includes RoleAssignments,RootFolder.Folders

                    foreach($PermissionObject in $ListPermissions.RoleAssignments)
                    {
                        #The usage of $RoleObject is not needed however hiding the Get-PnPProperty is.
                        $RoleObject = Get-PnPProperty -ClientObject $PermissionObject -Property Member, RoleDefinitionBindings

                        #Group or User name
                        $Name = $PermissionObject.Member.Title
        
                        #Full Control, Contribute, Edit etc.
                        $PermissionLevel = $PermissionObject.RoleDefinitionBindings | select -ExpandProperty Name
        
                        #Is it a SharePoint Group, Domain Group etc?
                        $ObjectType = $PermissionObject.Member.PrincipalType.ToString()

                        if($PermissionLevel -eq "Edit" -or $PermissionLevel -eq "Full Control" -or $PermissionLevel -eq "Read" -or $PermissionLevel -eq "Contribute"){
            
                            #Clean String before adding to array:
                            $Name += ","
                            $PermissionLevel += ","
                            #$ObjectType = $ObjectType.replace("`n","`r`n")

                            $script:CSVListPermissions += @("$Name $PermissionLevel $ObjectType")
                        }
                    }

                    #Create a blank line after the list is complete
                    $script:CSVListPermissions += @(",,")
                }
            }
        }
    }
    catch{
        Write-Host "`nGetting List Permissions failed..."
        Write-Host $_.Exception.Message
    }

    Write-Host "Finalising list permissions."
    try{
        $script:CSVListPermissions | Out-File $script:OutputFilePath -Append
    }
    catch{
        Write-Host -ForegroundColor Red "`n Outputting list permissions failed..."
        Write-Host $_.Exception.Message
    }
}

function GetSitePermissions{
 
    Write-Host "Getting Site Permissions"

    try{
        $web = Get-PnPWeb -Includes RoleAssignments

        #Add the header for this section of the final CSV:
        $script:SitePermissions += @("`nSite Permissions,Permission Level,Entity Type")


        foreach($PermissionObject in $web.RoleAssignments){
    
            $ObjectMember = Get-PnPProperty -ClientObject $PermissionObject -Property Member, RoleDefinitionBindings

            #Group or User name
            $Name = $PermissionObject.Member.Title
        
            #Full Control, Contribute, Edit etc.
            $PermissionLevel = $PermissionObject.RoleDefinitionBindings | select -ExpandProperty Name
        
            #Is it a SharePoint Group, Domain Group etc?
            $ObjectType = $PermissionObject.Member.PrincipalType.ToString()



            #We only want to log the users or groups that have direct permissions at the Site-Level, not limited access - limited access means it has permissions to the site else where, like a DLAG with edit perms on a document library
            if($PermissionLevel -eq "Edit" -or $PermissionLevel -eq "Full Control" -or $PermissionLevel -eq "Read" -or $PermissionLevel -eq "Contribute"){
            
                #Clean String before adding to array:
                $Name += ","
                $PermissionLevel += ","
                #$ObjectType = $ObjectType.replace("`n","`r`n")

                $script:SitePermissions += @("$Name $PermissionLevel $ObjectType")
            }
        }
    }
    catch{
        Write-Host -ForegroundColor Red "`nError Getting site permissions..."
        Write-Host $_.Exception.Message
    }

    Write-Host "Finalising Site Permissions"

    try{
        $script:SitePermissions | Out-File $script:OutputFilePath -Append
    }
    catch{
        Write-Host -ForegroundColor Red "`n Outputting site permissions failed..."
        Write-Host $_.Exception.Message
    }
}

function CreateCSV{
    try{
        #Get the name of the site and append to the output folder location to create the final output file path
        $Script:SiteName = Get-PnPWeb | Select -ExpandProperty Title
    
        $script:OutputFilePath = $script:OutputFolderLocation + "\$script:SiteName.csv"

        #If file already exists - would the user like to override it?
        if(Test-Path -Path "$script:OutputFilePath"){
            Write-Host -ForegroundColor Red "`nWarning: A file with the name '$script:SiteName' already exists within '$script:OutputFolderLocation'"
            $Choice = Read-Host "Would you like to override the file in this location ('Yes' or 'No')?"

            switch($Choice){
                Yes {break}
                No {StartScript;Start}
                Default {"That's not a valid answer!";CreateCSV;Break}
            }
        }

        $SiteURL = Get-PnPContext | Select -ExpandProperty Url
        $SiteVersion = Get-PnPContext | Select -ExpandProperty ServerLibraryVersion
        
        #$CSVOutput += @("Tenant Name, Site URL") 

        $CreateCSVObject = [PSCustomObject]@{
            "Site Name"    = $SiteName
            "Site URL"     = $SiteURL
            "Site Version" = $SiteVersion
        }
        $CreateCSVObject | Export-Csv "$script:OutputFilePath" -NoTypeInformation
    }
    catch{
        Write-Host -ForegroundColor Red "`nCreating CSV failed"
        Write-Host $_.Exception.Message
        exit
    }
}

function GetOutputFolder{
    
    #Open a interactive dialog for the user to select the output folder.
    try{
        [System.Reflection.Assembly]::LoadWithPartialName("System.windows.forms")|Out-Null
        $OutputFolderName = New-Object System.Windows.Forms.FolderBrowserDialog
        $OutputFolderName.Description = "Select a folder to output the Audit into"
        $OutputFolderName.RootFolder = "MyComputer"

        #I'm not sure why but if there's no variable to store ths .ShowDialog() the action get included within the $SelectedPath
        $catch = $OutputFolderName.ShowDialog()

        $SelectedPath = $OutputFolderName.SelectedPath

        #If the window is escaped or closed, ensure to reopen it otherwise the script will fail
        if($SelectedPath -eq $null -or $SelectedPath -eq ""){
            Write-Host -ForegroundColor Red "Select a folder to output the Audit into"
            GetOutputFolder
        }
        else{
            $folder = $OutputFolderName.SelectedPath.ToString()
        }
        return $folder
    }
    catch{
        Write-Host -ForegroundColor Red "`n Error getting output folder..."
        Write-Host $_.Exception.Message
    }
}

function LoginToSite{

    #Tidy the final output path of the "to be" csv
    $script:OutputFolderLocation = GetOutputFolder

    $SiteURL = Read-Host "`nPlease Enter the URL of the SharePoint site you want to audit`nEnsure this is the 'home' of the site (not a list URL)"

    try{
    #If the input is valid, connect and then call the GetSitePermissions function
        Connect-PnPOnline -Url "$SiteURL" -Interactive
        $script:SiteID = Get-PnPSite -Includes Id -ErrorVariable Loginerror | Select -ExpandProperty Id | Select -ExpandProperty Guid

        if($Loginerror -ne $null){
            throw "Something happened whilst Getting the Site information."
        }

    }
    catch{
        Write-Host -ForegroundColor Red "`nConnection failed, please ensure the URL is correct & the user has access to the site."
        Write-Host $_.Exception.Message
        
        Disconnect-PnPOnline
        LoginToSite
    }
}

function About{
    Clear-Host

    Write-Host "=================================================================================="
    Write-Host -ForegroundColor Yellow "Created by Jack Mumford 21/06/2021 Version 0.4"
    Write-Host ""
    Write-Host "This script will audit a selected SharePoint site and output the permissions of`n the site and lists (inc document libraries) into a destined CSV file."
    Write-Host "=================================================================================="
    Write-Host "Audit Types:"
    Write-Host "Simple Audit > Gets the top level of permissions for a site and its lists`n it also flags document libraries that have item with unique permissions"
    Write-Host "=================================================================================="

    Read-Host "Press ENTER to continue"
    Clear-Host
    StartScript
}

function Complete{
    #When the script has completed its output ask the user if they'd like to open the folder
    Write-Host -ForegroundColor Green -Object "Done!"
    $Choice = Read-Host "Would you like to open the audit CSV? ('Yes' or 'No')?"

    switch($Choice){
        Yes {Invoke-Item "$script:OutputFilePath";break}
        No {break}
        Default {"That's not a valid answer!";Complete;Break}
    }
    Disconnect-PnPOnline
    StartScript
}

function StartScript{
    Write-Host "Checking for PnP.PowerShell module..."
    $PnPModules = Get-Module -ListAvailable | where {$_.Name -eq "PnP.PowerShell"}

    if($PnPModules.Length -ge 1){
        #If a module with the name of PnP.Powershell is found:
        Write-Host "PnP.PowerShell module found!"
        Write-Host "================================"
        Write-Host -ForegroundColor Yellow "SharePoint Permissions Auditor"
        Write-Host "================================"
        Write-Host "1: Start Audit"
        Write-Host "2: About"
        Write-Host "3: Quit"
        Write-Host "================================"

        $StartScriptChoice = Read-Host "Please select a function (1,2,3)"

        switch($StartScriptChoice){
            1 {
                LoginToSite
                CreateCSV
                GetSitePermissions
                GetListPermissions
                Complete
                break
            }
            2 {About;break}
            3 {exit;break}
            default{Clear-Host;StartScript}
        }
    }
    else{
        try{
            Install-Module -Name PnP.Powershell -Force
        }
        catch{
            Write-Host -ForegroundColor Red "Something went wrong installing the PnP.PowerShell module, please ensure you install this before running again."
            Write-Host $_.Exception.Message
            exit
        }
        Write-Host -ForegroundColor DarkGreen -Object "Installed PnP.PowerShell Module"
        StartScript
    }
}

StartScript
Disconnect-PnPOnline